const p = s => 'exnihilo:' + s
const defaultBlockTag = p('sieve')
const blockProp = s => `q.block_property('${s}')`
const tbP = s => blockProp(p(s))

function prefix(s: string, minecraft = false) {
    return `${minecraft ? "minecraft" : this.namespace_prefix}:${s}`
}

function is_item_name_any(items, location = "slot.weapon.mainhand", slotId = 0) {
    let item_string = ""
    items.every((v, i) => {
        item_string += `'${v}'`
        item_string += (i === items.length - 1 ? "" : ", ")
    })
    return `q.is_item_name_any('${location}', ${slotId}, ${item_string})`
}

const join = (l: any[]) => l.join("/")
const namespace_prefix = "exnihilo"
const block_tag = prefix("sieve")
const block_property = (name: string) => `q.block_property('${prefix(name)}')`
const geometry = (name: string) => "geometry." + name
const block = prefix("block", true)
const description = join([block, "description"])
const components = join([block, "components"])
const events = join([block, "events"])
const permutations = join([block, "permutations"])
const new_event = (name: string) => join([events, name])
const set_block_property = (name: string[], value: any[]) => {
    let r = { set_block_property: {} }
    name.forEach((v, i) => {
        r.set_block_property[v] = value[i]
    })
    return r
}

export class Sieve {
    readonly schema: Function
    readonly template: Function

    constructor({ name, schema, template }) {
        name(block_tag)

        this.schema = schema
        this.template = template

        this.template((s, { create }: any) => [
            this.setup_block_tag,
            this.setup_default_block_geometry,
            this.setup_default_material_instances,
            this.setup_destructible,
            this.setup_part_visibility,
            this.setup_events,
        ].forEach(f => f.call({ s, create })))
    }

    private setup_block_tag({ create }) {
        create({
            [`tag:${block_tag}`]: {}
        }, components)
    }

    private setup_default_block_geometry({ create }) {
        create({
            [prefix("geometry", true)]: geometry("sieve")
        }, components)
    }

    private setup_default_material_instances({ s, create }) {
        create({
            '*': {
                texture: `stripped_${s.wood_type}_log_side`,
                render_method: 'alpha_test'
            }
        }, join([components, prefix("material_instances", true)]))
    }

    private setup_destructible({ create }) {
        create({
            [prefix("destructible_by_explosion", true)]: {
                explosion_resistance: 3.0
            }
        }, components)

        create({
            [prefix("destructible_by_mining", true)]: {
                seconds_to_destroy: 3.0
            }
        }, components)
    }

    private setup_part_visibility({ create }) {
        let path = join([components, prefix("part_visility"), "conditions"])
        create({
            mesh: block_property("mesh"),
            material: block_property("material"),
            bind: block_property("stage"),
        }, path)

        for (let i = 1; i <= 10; i++) {
            create({
                [`stage${i}`]: block_property("stage") + "==" + i
            }, path)
        }
    }

    private setup_events(p) {
        this.event_reset_mesh(p)
        this.event_reset_material(p)
        this.event_empty_interact(p)
    }

    private event_reset_mesh({ create }) {
        create(set_block_property(["mesh"], [0]), new_event("reset_mesh"))
    }

    private event_reset_material({ create }) {
        create(set_block_property(["material", "stage"], [0, 0]), 'minecraft:block/events/reset_material')
    }

    private event_empty_interact({ s, create }) {
        let e = { sequence: [] }
        s.mesh_list.forEach((i: string, i2: number) => {
            e.sequence.push({
                condition: is_item_name_any(prefix("mesh_" + i)),
                decrement_stack: {
                    ignore_game_mode: true
                }
            })
        })
        create(e, new_event("empty"))
    }
}

export default defineComponent(p => {
    new Sieve(p)
})

defineComponent(({ name, schema, template }) => {
    name(p('sieve'))
    schema({
        description: 'Make this block is a sieve',
        type: 'object',
        anyOf: [
            { required: ['wood_type', 'textures', 'material_list', 'mesh_list'] }
        ],
        properties: {
            wood_type: {
                description: 'Define this block material of wood type',
                type: 'string',
                enum: ['oak', 'spruce', 'birch', 'jungle', 'acacia', 'dark_oak', 'mangrove']
            },
            textures: {
                description: 'Define frame texture',
                type: 'string'
            },
            material_list: {
                description: 'Define material type this sieve can put in for sieve',
                type: 'array'
            },
            mesh_list: {
                description: 'Define mesh type this sieve can have\n["", "", "",... etc]',
                type: 'array'
            },
            type_data: {
                description: 'Additional type data for every material can sieve as numberic\nmaterial_list: [ "dirt" ],\ntype_data: {\n"dirt": [ "normal", "coarse" ]\n}',
                type: 'object'
            }
        }
    })

    template(({ wood_type, textures, material_list, mesh_list, type_data }, { create }) => {

        let e: { sequence: any }

        e = {
            sequence: [{
                condition: `q.is_sneaking && !${tbP('meterial')}`,
                trigger: 'drop_mesh'
            }]
        }
        create(e, 'minecraft:block/events/mesh')

        e = { sequence: [] }
        mesh_list.forEach((i: string, i2: number) => {
            e.sequence.push({
                condition: `${tbP('mesh')}==${i2 + 1}`,
                spawn_loot: {
                    table: `loot_tables/mesh/${i}.json`
                },
                trigger: 'reset_mesh'
            })
        })
        create(e, 'minecraft:block/events/drop_mesh')

        mesh_list.forEach((i: string, i2: number) => {
            create({
                condition: `${tbP('mesh')}==${i2 + 1}`,
                components: {
                    "minecraft:material_instances": {
                        "*": {
                            texture: textures,
                            render_method: "alpha_test"
                        },
                        "up": {
                            texture: `mesh_${i}`,
                            render_method: "alpha_test"
                        }
                    }
                }
            }, 'minecraft:block/permutations')
        })
    })
})
