import { Block, BlockLocation, Entity, ItemStack, world } from "@minecraft/server";
import { BlockBase } from "./block_base";
import { logerr, prefix, splitNamespace } from "../utils";

export class Sieve extends BlockBase {

    static readonly NAMESPACE = prefix("sieve");

    static isSieveBlock(block: Block) {
        return block.hasTag(Sieve.NAMESPACE);
    }

    static Registry() {
        world.events.itemUseOn.subscribe(({ blockLocation, item, source }) => {
            if (source.isSneaking) return;
            let block = source.dimension.getBlock(blockLocation);
            if (!Sieve.isSieveBlock(block)) return;
            (({ x, y, z }) => {
                let blocks = new BlockLocation(x + 2, y, z + 2).blocksBetween(new BlockLocation(x - 2, y, z - 2));
                blocks.forEach((location) => {
                    let block = source.dimension.getBlock(location);
                    if (Sieve.isSieveBlock(block)) new Sieve(block).onInteract(source, item);
                })
            }).call(this, blockLocation);
        })
    }

    public onInteract(source: Entity, item: ItemStack) {
        let mesh = this.getBlockProperty("mesh") as number;
        if (!mesh) return;
        let stage = this.getBlockProperty("stage") as number;
        if (stage > 0) {
            if (stage == 10) {
                let loot = "";

                let main_type = this.getBlockProperty("main_type") as number;
                let sub_type = this.getBlockProperty("sub_type") as number;

                switch (main_type) {
                    // none
                    case 0:
                        break;
                    // gravel
                    case 1:
                        if (mesh > 3) loot = "gravel/4";
                        else loot = "gravel/" + mesh;
                        break;
                    // leaves
                    case 2:
                        let leaves_type = [
                            "oak", "spruce", "birch", "jungle",
                            "acacia", "dark_oak", "mangrove",
                            "azalea", "azalea_flowered"
                        ];
                        loot = "leaves/" + leaves_type[sub_type];
                        if (mesh > 3) loot += 4;
                        else loot += mesh;
                        break;
                    // crushed
                    case 3:
                        if (sub_type == 0) {
                            if (mesh > 3) loot = "crushed/netherrack/4";
                            else loot = "crushed/netherrack/" + mesh;
                        } else if (sub_type == 1) {
                            if (mesh < 3) loot = "crushed/end_stone/1";
                            else loot = "crushed/end_stone/" + (mesh - 1);
                        }
                        break;
                    // sand
                    case 4:
                        break;
                    // dirt
                    case 5:
                        loot = "dirt";
                        if (sub_type == 1) loot = "coarse_dirt";
                        break;
                    // dust
                    case 6:
                        if (mesh < 3) loot = "dust/1";
                        else loot = "dust/2";
                        break;
                }

                this.setBlockProperties({
                    "main_type": 0,
                    "sub_type": 0,
                    "stage": 0
                });

                if (loot) {
                    let entity = this.block.dimension.spawnEntity(prefix("drop_item_entity"), this.location);
                    entity.runCommandAsync(`loot replace entity @s slot.inventory 0 loot "sieve/${loot}"`)
                        .then(() => {
                            // @ts-ignored
                            let container = entity.getComponent("inventory").container;
                            for (let i = 0; i < container.size - container.emptySlotsCount; i++) {
                                this.block.dimension.spawnItem(container.getItem(i), this.location.above());
                            }
                        })
                        .finally(() => entity.triggerEvent("despawn"));
                }
            } else {
                this.setBlockProperties({ "stage": stage + 1 });
            }
        } else {
            let splited = splitNamespace(item.typeId);

            const clearItem = () => {
                source.runCommandAsync(`clear @s[m!=c] ${item.typeId} ${item.data} 1`);
            };

            const setBlockProperties = (properties: { [name: string]: string | number | boolean }) => {
                properties["stage"] = 1;
                clearItem();
                this.setBlockProperties(properties);
            };

            if (splited.isVanilla) {
                switch (splited.last) {
                    case "dirt":
                        setBlockProperties({
                            "main_type": 5,
                            "sub_type": item.data
                        });
                        break;
                    case "sand":
                        setBlockProperties({
                            "main_type": 4,
                            "sub_type": item.data
                        });
                        break;
                    case "soul_sand":
                        setBlockProperties({
                            "main_type": 4,
                            "sub_type": 3
                        });
                        break;
                    case "gravel":
                        setBlockProperties({
                            "main_type": 1
                        });
                        break;
                    case "leaves":
                        setBlockProperties({
                            "main_type": 2,
                            "sub_type": item.data
                        });
                        break;
                    case "leaves2":
                        setBlockProperties({
                            "main_type": 2,
                            "sub_type": item.data + 4
                        });
                        break;
                    case "mangrove_leaves":
                        setBlockProperties({
                            "main_type": 2,
                            "sub_type": 6
                        });
                        break;
                    case "azalea_leaves":
                        setBlockProperties({
                            "main_type": 2,
                            "sub_type": 7
                        });
                        break;
                    case "azalea_leaves_flowered":
                        setBlockProperties({
                            "main_type": 2,
                            "sub_type": 8
                        });
                        break;
                }
            } else {
                switch (splited.first) {
                    case "exnihilo":
                        switch (splited.last) {
                            case "crushed_netherrack":
                                setBlockProperties({
                                    "main_type": 3,
                                    "sub_type": 0
                                });
                                break;
                            case "crushed_end_stone":
                                setBlockProperties({
                                    "main_type": 3,
                                    "sub_type": 1
                                });
                                break;
                        }
                        break;
                }
            }
        }
    }
}